from pathlib import Path
import torch
import torch.nn.functional as F
from .model.model import ConcordModel
from .utils.preprocessor import Preprocessor
from .utils.anndata_utils import ensure_categorical
from .model.dataloader import DataLoaderManager 
from .model.chunkloader import ChunkLoader
from .utils.other_util import add_file_handler, set_seed
from .model.trainer import Trainer
import numpy as np
import scanpy as sc
import pandas as pd
import copy
import json
from . import logger
from . import set_verbose_mode

class Config:
    def __init__(self, config_dict):
        for key, value in config_dict.items():
            setattr(self, key, value)

    def __getitem__(self, item):
        return getattr(self, item)

    def to_dict(self):
        def serialize(value):
            if isinstance(value, (torch.device,)):
                return str(value)
            # Add more cases if needed for other non-serializable types
            return value

        return {key: serialize(getattr(self, key)) for key in dir(self)
                if not key.startswith('__') and not callable(getattr(self, key))}


class Concord:
    """
    A contrastive learning framework for single-cell data analysis.

    CONCORD performs dimensionality reduction, denoising, and batch correction 
    in an unsupervised manner while preserving local and global topological structures.

    Attributes:
        adata (AnnData): Input AnnData object.
        save_dir (Path): Directory to save outputs and logs.
        config (Config): Configuration object storing hyperparameters.
        model (ConcordModel): The main contrastive learning model.
        trainer (Trainer): Handles model training.
        loader (DataLoaderManager or ChunkLoader): Data loading utilities.
    """
    def __init__(self, adata, save_dir='save/', inplace=True, verbose=False, **kwargs):
        """
        Initializes the Concord framework.

        Args:
            adata (AnnData): Input single-cell data in AnnData format.
            save_dir (str, optional): Directory to save model outputs. Defaults to 'save/'.
            inplace (bool, optional): If True, modifies `adata` in place. Defaults to True.
            verbose (bool, optional): Enable verbose logging. Defaults to False.
            **kwargs: Additional configuration parameters.

        Raises:
            ValueError: If `inplace` is set to True on a backed AnnData object.
        """
        set_verbose_mode(verbose)
        if adata.isbacked:
            logger.warning("Input AnnData object is backed. With same amount of epochs, Concord will perform better when adata is loaded into memory.")
            if inplace:
                raise ValueError("Inplace mode is not supported for backed AnnData object. Set inplace to False.")
            self.adata = adata
        else:
            self.adata = adata if inplace else adata.copy()

        self.save_dir = Path(save_dir)
        self.config = None
        self.loader = None
        self.model = None
        self.run = None
        self.sampler_kwargs = {}

        if not self.save_dir.exists():
            self.save_dir.mkdir(parents=True, exist_ok=True)

        add_file_handler(logger, self.save_dir / "run.log")

        self.default_params = dict(
            seed=0,
            project_name="concord",
            input_feature=None,
            batch_size=64,
            n_epochs=10,
            lr=1e-2,
            schedule_ratio=0.97,
            train_frac=1.0,
            latent_dim=100,
            encoder_dims=[512],
            decoder_dims=[512],
            augmentation_mask_prob=0.4,  
            domain_key=None,
            class_key=None,
            domain_embedding_dim=8,
            covariate_embedding_dims={},
            use_decoder=False, # Default decoder usage
            decoder_final_activation='relu',
            decoder_weight=1.0,
            clr_mode="aug", # Consider fix
            clr_temperature=0.5,
            clr_weight=1.0,
            use_classifier=False,
            classifier_weight=1.0,
            unlabeled_class=None,
            use_importance_mask=False,
            importance_penalty_weight=0,
            importance_penalty_type='L1',
            dropout_prob=0.1,
            norm_type="layer_norm",  # Default normalization type
            sampler_emb="X_pca",
            sampler_knn=None, # Default neighborhood size, can be adjusted
            dist_metric='euclidean',
            p_intra_knn=0.3,
            p_intra_domain=0.95,
            min_p_intra_domain=0.9,
            max_p_intra_domain=1.0,
            pca_n_comps=50,
            use_faiss=True,
            use_ivf=True,
            ivf_nprobe=10,
            pretrained_model=None,
            classifier_freeze_param=False,
            chunked=False,
            chunk_size=10000,
            #encoder_append_cov=False, # Should always be False for now
            device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        )

        self.setup_config(**kwargs)
        set_seed(self.config.seed)

        if self.config.sampler_knn is None:
            self.config.sampler_knn = self.adata.n_obs // 10 
            logger.info(f"Setting sampler_knn to {self.config.sampler_knn} to be 1/10 the number of cells in the dataset. You can change this value by setting sampler_knn in the configuration.")

        # Checks to convert None values to default values
        if self.config.input_feature is None:
            logger.warning("No input feature list provided. It is recommended to first select features using the command `concord.ul.select_features()`.")
            logger.info(f"Proceeding with all {self.adata.shape[1]} features in the dataset.")
            self.config.input_feature = self.adata.var_names.tolist()

        if self.config.importance_penalty_weight == 0 and self.config.use_importance_mask:
            logger.warning("Importance mask is enabled but importance_penalty_weight is set to 0.0. This will still cause differential weighting of features, but without penalty.")

        if self.config.domain_key is not None:
            if(self.config.domain_key not in self.adata.obs.columns):
                raise ValueError(f"Domain key {self.config.domain_key} not found in adata.obs. Please provide a valid domain key.")
            ensure_categorical(self.adata, obs_key=self.config.domain_key, drop_unused=True)
        else:
            logger.warning("domain/batch information not found, all samples will be treated as from single domain/batch.")
            self.config.domain_key = 'tmp_domain_label'
            self.adata.obs[self.config.domain_key] = pd.Series(data='single_domain', index=self.adata.obs_names).astype('category')
            self.p_intra_domain = 1.0

        self.num_domains = len(self.adata.obs[self.config.domain_key].cat.categories)

        # User must set p_intra_domain = 1.0 or min_p_intra_domain to 1.0 to use this feature
        # if self.config.encoder_append_cov:
        #     if self.config.p_intra_domain != 1.0:
        #         if self.config.min_p_intra_domain != 1.0:
        #             raise ValueError("User must set p_intra_domain = 1.0 when encoder_append_cov is True, otherwise set it to False.")

        if self.config.train_frac < 1.0 and self.config.p_intra_knn > 0:
            logger.warning("Nearest neighbor contrastive loss is currently not supported for training fraction less than 1.0. Setting p_intra_knn to 0.")
            self.config.p_intra_knn = 0

        # Check if batch_size conflicts with p_intra_knn
        batch_knn_count = int(self.config.p_intra_knn * self.config.batch_size)
        if self.config.p_intra_knn > 0 and batch_knn_count > self.config.sampler_knn:
            logger.warning(f"p_intra_knn * batch_size ({batch_knn_count}) is greater than sampler_knn ({self.config.sampler_knn}). This will cause actual sampling ratio not matching specified p_intra_knn.")
            logger.warning(f"You should either set batch_size to be smaller than sampler_knn/p_intra_knn ({int(self.config.sampler_knn/self.config.p_intra_knn)})")
            logger.warning(f"or set sampler_knn to be greater than p_intra_knn * batch_size ({batch_knn_count}).")
            #self.config.p_intra_knn = 0

        if self.config.use_classifier:
            if self.config.class_key is None:
                raise ValueError("Cannot use classifier without providing a class key.")
            if(self.config.class_key not in self.adata.obs.columns):
                raise ValueError(f"Class key {self.config.class_key} not found in adata.obs. Please provide a valid class key.")
            ensure_categorical(self.adata, obs_key=self.config.class_key, drop_unused=True)

            self.unique_classes = self.adata.obs[self.config.class_key].cat.categories
            self.unique_classes_code = [code for code, _ in enumerate(self.unique_classes)]
            if self.config.unlabeled_class is not None:
                if self.config.unlabeled_class in self.unique_classes:
                    self.unlabeled_class_code = self.unique_classes.get_loc(self.config.unlabeled_class)
                    self.unique_classes_code = self.unique_classes_code[self.unique_classes_code != self.unlabeled_class_code]
                    self.unique_classes = self.unique_classes[self.unique_classes != self.config.unlabeled_class]
                else:
                    raise ValueError(f"Unlabeled class {self.config.unlabeled_class} not found in the class key.")
            else:
                self.unlabeled_class_code = None

            self.num_classes = len(self.unique_classes_code)
        else:
            self.unique_classes = None
            self.unique_classes_code = None
            self.unlabeled_class_code = None
            self.num_classes = None

        # Compute the number of categories for each covariate
        self.covariate_num_categories = {}
        for covariate_key in self.config.covariate_embedding_dims.keys():
            if covariate_key in self.adata.obs:
                ensure_categorical(self.adata, obs_key=covariate_key, drop_unused=True)
                self.covariate_num_categories[covariate_key] = len(self.adata.obs[covariate_key].cat.categories)
        
        # to be used by chunkloader for data transformation
        self.preprocessor = Preprocessor(
            use_key="X",
            feature_list=self.config.input_feature,
            normalize_total=1e4,
            result_normed_key="X_normed",
            log1p=True,
            result_log1p_key="X_log1p"
        )


    def get_default_params(self):
        """
        Returns the default hyperparameters used in CONCORD.

        Returns:
            dict: A dictionary containing default configuration values.
        """
        return self.default_params.copy()
    

    def setup_config(self, **kwargs):
        """
        Sets up the configuration for training.

        Args:
            **kwargs: Key-value pairs to override default parameters.

        Raises:
            ValueError: If an invalid parameter is provided.
        """
        # Start with the default parameters
        initial_params = self.default_params.copy()

        # Check if any of the provided parameters are not in the default parameters
        invalid_params = set(kwargs.keys()) - set(initial_params.keys())
        if invalid_params:
            raise ValueError(f"Invalid parameters provided: {invalid_params}")

        # Update with user-provided values (if any)
        initial_params.update(kwargs)

        self.config = Config(initial_params)


    def init_model(self):
        """
        Initializes the CONCORD model and loads a pre-trained model if specified.

        Raises:
            FileNotFoundError: If the specified pre-trained model file is missing.
        """
        input_dim = len(self.config.input_feature)
        hidden_dim = self.config.latent_dim

        self.model = ConcordModel(input_dim, hidden_dim, 
                                  num_domains=self.num_domains,
                                  num_classes=self.num_classes,
                                  domain_embedding_dim=self.config.domain_embedding_dim,
                                  covariate_embedding_dims=self.config.covariate_embedding_dims,
                                  covariate_num_categories=self.covariate_num_categories,
                                  #encoder_append_cov=self.config.encoder_append_cov,
                                  encoder_dims=self.config.encoder_dims,
                                  decoder_dims=self.config.decoder_dims,
                                  decoder_final_activation=self.config.decoder_final_activation,
                                  augmentation_mask_prob=self.config.augmentation_mask_prob,
                                  dropout_prob=self.config.dropout_prob,
                                  norm_type=self.config.norm_type,
                                  use_decoder=self.config.use_decoder,
                                  use_classifier=self.config.use_classifier,
                                  use_importance_mask=self.config.use_importance_mask).to(self.config.device)

        logger.info(f'Model loaded to device: {self.config.device}')
        total_params = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
        logger.info(f'Total number of parameters: {total_params}')

        if self.config.pretrained_model is not None:
            pretrained_model_path = Path(self.config.pretrained_model)
            if pretrained_model_path.exists():
                logger.info(f"Loading pre-trained model from {pretrained_model_path}")
                self.model.load_model(pretrained_model_path, self.config.device)
            else:
                raise FileNotFoundError(f"Model file not found at {pretrained_model_path}")
            

    def init_trainer(self):
        """
        Initializes the model trainer, setting up loss functions, optimizer, and learning rate scheduler.
        """
        self.trainer = Trainer(model=self.model,
                               data_structure=self.data_structure,
                               device=self.config.device,
                               logger=logger,
                               lr=self.config.lr,
                               schedule_ratio=self.config.schedule_ratio,
                               use_classifier=self.config.use_classifier, 
                               classifier_weight=self.config.classifier_weight,
                               unique_classes=self.unique_classes_code,
                               unlabeled_class=self.unlabeled_class_code,
                               use_decoder=self.config.use_decoder,
                               decoder_weight=self.config.decoder_weight,
                               clr_mode=self.config.clr_mode, 
                               clr_temperature=self.config.clr_temperature,
                               clr_weight=self.config.clr_weight,
                               importance_penalty_weight=self.config.importance_penalty_weight,
                               importance_penalty_type=self.config.importance_penalty_type)


    def init_dataloader(self, input_layer_key='X_log1p', preprocess=True, train_frac=1.0, use_sampler=True):
        """
        Initializes the data loader for training and evaluation.

        Args:
            input_layer_key (str, optional): Key in `adata.layers` to use as input. Defaults to 'X_log1p'.
            preprocess (bool, optional): Whether to apply preprocessing. Defaults to True.
            train_frac (float, optional): Fraction of data to use for training. Defaults to 1.0.
            use_sampler (bool, optional): Whether to use the probabilistic sampler. Defaults to True.

        Raises:
            ValueError: If `train_frac < 1.0` and contrastive loss mode is 'nn'.
        """
        if train_frac < 1.0 and self.config.clr_mode == 'nn':
            raise ValueError("Nearest neighbor contrastive loss is not supported for training fraction less than 1.0.")
        self.data_manager = DataLoaderManager(
            input_layer_key=input_layer_key, domain_key=self.config.domain_key, 
            class_key=self.config.class_key, covariate_keys=self.config.covariate_embedding_dims.keys(), 
            batch_size=self.config.batch_size, train_frac=train_frac,
            use_sampler=use_sampler,
            sampler_emb=self.config.sampler_emb, 
            sampler_knn=self.config.sampler_knn,
            dist_metric=self.config.dist_metric, 
            p_intra_knn=self.config.p_intra_knn, 
            p_intra_domain=self.config.p_intra_domain, 
            min_p_intra_domain=self.config.min_p_intra_domain,
            max_p_intra_domain=self.config.max_p_intra_domain,
            clr_mode=self.config.clr_mode, 
            pca_n_comps=self.config.pca_n_comps,
            use_faiss=self.config.use_faiss, 
            use_ivf=self.config.use_ivf, 
            ivf_nprobe=self.config.ivf_nprobe, 
            preprocess=self.preprocessor if preprocess else None,
            num_cores=self.num_classes, 
            device=self.config.device
        )

        if self.config.chunked:
            self.loader = ChunkLoader(
                adata=self.adata,
                chunk_size=self.config.chunk_size,
                data_manager=self.data_manager
            )
            self.data_structure = self.loader.data_structure  # Retrieve data_structure
        else:
            train_dataloader, val_dataloader, self.data_structure = self.data_manager.anndata_to_dataloader(self.adata)
            self.loader = [(train_dataloader, val_dataloader, np.arange(self.adata.shape[0]))]


    def train(self, save_model=True, patience=2):
        """
        Trains the model on the dataset.

        Args:
            save_model (bool, optional): Whether to save the trained model. Defaults to True.
            patience (int, optional): Number of epochs to wait for improvement before early stopping. Defaults to 2.
        """
        best_val_loss = float('inf')
        best_model_state = None
        epochs_without_improvement = 0

        for epoch in range(self.config.n_epochs):
            logger.info(f'Starting epoch {epoch + 1}/{self.config.n_epochs}')
            for chunk_idx, (train_dataloader, val_dataloader, _) in enumerate(self.loader):
                logger.info(f'Processing chunk {chunk_idx + 1}/{len(self.loader)} for epoch {epoch + 1}')
                if train_dataloader is not None:
                    logger.info(f"Number of samples in train_dataloader: {len(train_dataloader.dataset)}")
                if val_dataloader is not None:
                    logger.info(f"Number of samples in val_dataloader: {len(val_dataloader.dataset)}")

                self.trainer.train_epoch(epoch, train_dataloader)
                
                if val_dataloader is not None:
                    val_loss = self.trainer.validate_epoch(epoch, val_dataloader)
                
                    # Check if the current validation loss is the best we've seen so far
                    if val_loss < best_val_loss:
                        best_val_loss = val_loss
                        best_model_state = copy.deepcopy(self.model.state_dict())
                        logger.info(f"New best model found at epoch {epoch + 1} with validation loss: {best_val_loss:.4f}")
                        epochs_without_improvement = 0  # Reset counter when improvement is found
                    else:
                        epochs_without_improvement += 1
                        logger.info(f"No improvement in validation loss for {epochs_without_improvement} epoch(s).")

                    # Early stopping condition
                    if epochs_without_improvement >= patience:
                        logger.info(f"Stopping early at epoch {epoch + 1} due to no improvement in validation loss.")
                        break

            self.trainer.scheduler.step()

            # Early stopping break condition
            if epochs_without_improvement > patience:
                break

        if best_model_state is not None:
            self.model.load_state_dict(best_model_state)
            logger.info("Best model state loaded into the model before final save.")

        if save_model:
            import time
            file_suffix = f"{time.strftime('%b%d-%H%M')}"
            model_save_path = self.save_dir / f"final_model_{file_suffix}.pt"
            self.save_model(self.model, model_save_path)
            # Save the configuration
            config_save_path = self.save_dir / f"config_{file_suffix}.json"
            with open(config_save_path, 'w') as f:
                json.dump(self.config.to_dict(), f, indent=4)
            
            logger.info(f"Final model saved at: {model_save_path}; Configuration saved at: {config_save_path}.")


    def predict(self, loader, sort_by_indices=False, return_decoded=False, decoder_domain=None, return_latent=False, return_class=True, return_class_prob=True):  
        """
        Runs inference on a dataset.

        Args:
            loader (DataLoader or list): Data loader or chunked loader for batch processing.
            sort_by_indices (bool, optional): Whether to return results in original cell order. Defaults to False.
            return_decoded (bool, optional): Whether to return decoded gene expression. Defaults to False.
            decoder_domain (str, optional): Specifies a domain for decoding. Defaults to None.
            return_latent (bool, optional): Whether to return latent variables. Defaults to False.
            return_class (bool, optional): Whether to return predicted class labels. Defaults to True.
            return_class_prob (bool, optional): Whether to return class probabilities. Defaults to True.

        Returns:
            tuple: Encoded embeddings, decoded matrix (if requested), class predictions, class probabilities, true labels, and latent variables.
        """
        self.model.eval()
        class_preds = []
        class_true = []
        class_probs = [] if return_class_prob else None
        embeddings = []
        decoded_mtx = []
        indices = []

        latent_matrices = {}
        
        if isinstance(loader, list) or type(loader).__name__ == 'ChunkLoader':
            all_embeddings = []
            all_decoded = []
            all_class_preds = []
            all_class_probs = [] if return_class_prob else None
            all_class_true = []
            all_indices = []
            all_latent_matrices = {}

            for chunk_idx, (dataloader, _, ck_indices) in enumerate(loader):
                logger.info(f'Predicting for chunk {chunk_idx + 1}/{len(loader)}')
                ck_embeddings, ck_decoded, ck_class_preds, ck_class_probs, ck_class_true, ck_latent = self.predict(dataloader, 
                                                                                        sort_by_indices=True, 
                                                                                        return_decoded=return_decoded, 
                                                                                        decoder_domain=decoder_domain,
                                                                                        return_latent=return_latent,
                                                                                        return_class=return_class,
                                                                                        return_class_prob=return_class_prob)
                all_embeddings.append(ck_embeddings)
                all_decoded.append(ck_decoded) if return_decoded else None
                all_indices.extend(ck_indices)
                if ck_class_preds is not None:
                    all_class_preds.extend(ck_class_preds)
                if return_class_prob and ck_class_probs is not None:
                    all_class_probs.append(ck_class_probs)
                if ck_class_true is not None:
                    all_class_true.extend(ck_class_true)
                if return_latent:
                    for key in ck_latent.keys():
                        if key not in all_latent_matrices:
                            all_latent_matrices[key] = []
                        all_latent_matrices[key].append(ck_latent[key])

            all_indices = np.array(all_indices)
            sorted_indices = np.argsort(all_indices)

            all_embeddings = np.concatenate(all_embeddings, axis=0)[sorted_indices]
            all_decoded = np.concatenate(all_decoded, axis=0)[sorted_indices] if all_decoded else None
            all_class_preds = np.array(all_class_preds)[sorted_indices] if all_class_preds else None
            all_class_true = np.array(all_class_true)[sorted_indices] if all_class_true else None
            if return_class_prob:
                all_class_probs = pd.concat(all_class_probs).iloc[sorted_indices].reset_index(drop=True) if all_class_probs else None
            
            if return_latent:
                for key in all_latent_matrices.keys():
                    all_latent_matrices[key] = np.concatenate(all_latent_matrices[key], axis=0)[sorted_indices]
            return all_embeddings, all_decoded, all_class_preds, all_class_probs, all_class_true, all_latent_matrices
        else:
            with torch.no_grad():
                if decoder_domain is not None:
                    logger.info(f"Projecting data back to expression space of specified domain: {decoder_domain}")
                    # map domain to actual domain id used in the model
                    fixed_domain_id = torch.tensor([self.adata.obs[self.config.domain_key].cat.categories.get_loc(decoder_domain)], dtype=torch.long).to(
                        self.config.device)
                else:
                    if self.config.use_decoder:
                        logger.info("No domain specified for decoding. Using the same domain as the input data.")
                    fixed_domain_id = None
                
                for data in loader:
                    # Unpack data based on the provided structure
                    data_dict = {key: value.to(self.config.device) for key, value in zip(self.data_structure, data)}

                    inputs = data_dict.get('input')
                    # Use fixed domain id if provided, and make it same length as inputs
                    domain_ids = data_dict.get('domain', None) if decoder_domain is None else fixed_domain_id.repeat(inputs.size(0))
                    class_labels = data_dict.get('class', None)
                    original_indices = data_dict.get('idx')
                    covariate_keys = [key for key in data_dict.keys() if key not in ['input', 'domain', 'class', 'idx']]
                    covariate_tensors = {key: data_dict[key] for key in covariate_keys}

                    if class_labels is not None:
                        class_true.extend(class_labels.cpu().numpy())

                    if original_indices is not None:
                        indices.extend(original_indices.cpu().numpy())

                    outputs = self.model(inputs, domain_ids, covariate_tensors, return_latent=return_latent)
                    if 'class_pred' in outputs and return_class:
                        class_preds_tensor = outputs['class_pred']
                        class_preds.extend(torch.argmax(class_preds_tensor, dim=1).cpu().numpy()) # TODO May need fix
                        if return_class_prob:
                            class_probs.extend(F.softmax(class_preds_tensor, dim=1).cpu().numpy())
                    if 'encoded' in outputs:
                        embeddings.append(outputs['encoded'].cpu().numpy())
                    if 'decoded' in outputs and return_decoded:
                        decoded_mtx.append(outputs['decoded'].cpu().numpy())
                    if 'latent' in outputs and return_latent:
                        latent = outputs['latent']
                        for key, val in latent.items():
                            if key not in latent_matrices:
                                latent_matrices[key] = []
                            latent_matrices[key].append(val.cpu().numpy())   
                            

            if not embeddings:
                raise ValueError("No embeddings were extracted. Check the model and dataloader.")

            # Concatenate embeddings
            embeddings = np.concatenate(embeddings, axis=0)

            if decoded_mtx:
                decoded_mtx = np.concatenate(decoded_mtx, axis=0)

            if return_latent:
                for key in latent_matrices.keys():
                    latent_matrices[key] = np.concatenate(latent_matrices[key], axis=0)

            # Convert predictions and true labels to numpy arrays
            class_preds = np.array(class_preds) if class_preds else None
            class_probs = np.array(class_probs) if return_class_prob and class_probs else None
            class_true = np.array(class_true) if class_true else None

            if sort_by_indices and indices:
                # Sort embeddings and predictions back to the original order
                indices = np.array(indices)
                sorted_indices = np.argsort(indices)
                embeddings = embeddings[sorted_indices]
                if return_decoded:
                    decoded_mtx = decoded_mtx[sorted_indices]
                if class_preds is not None:
                    class_preds = class_preds[sorted_indices]
                if return_class_prob and class_probs is not None:
                    class_probs = class_probs[sorted_indices]
                if class_true is not None:
                    class_true = class_true[sorted_indices]
                if return_latent:
                    for key in latent_matrices.keys():
                        latent_matrices[key] = latent_matrices[key][sorted_indices]

            if return_class and self.unique_classes is not None:
                class_preds = self.unique_classes[class_preds] if class_preds is not None else None
                class_true = self.unique_classes[class_true] if class_true is not None else None
                if return_class_prob and class_probs is not None:
                    class_probs = pd.DataFrame(class_probs, columns=self.unique_classes)

            return embeddings, decoded_mtx, class_preds, class_probs, class_true, latent_matrices


    def encode_adata(self, input_layer_key="X_log1p", output_key="Concord", preprocess=True, 
                     return_decoded=False, decoder_domain=None,
                     return_latent=False, 
                     return_class=True, return_class_prob=True, 
                     save_model=True):
        """
        Encodes an AnnData object using the CONCORD model.

        Args:
            input_layer_key (str, optional): Input layer key. Defaults to 'X_log1p'.
            output_key (str, optional): Output key for storing results in AnnData. Defaults to 'Concord'.
            preprocess (bool, optional): Whether to apply preprocessing. Defaults to True.
            return_decoded (bool, optional): Whether to return decoded gene expression. Defaults to False.
            decoder_domain (str, optional): Specifies domain for decoding. Defaults to None.
            return_latent (bool, optional): Whether to return latent variables. Defaults to False.
            return_class (bool, optional): Whether to return predicted class labels. Defaults to True.
            return_class_prob (bool, optional): Whether to return class probabilities. Defaults to True.
            save_model (bool, optional): Whether to save the model after training. Defaults to True.
        """

        # Initialize the model
        self.init_model()
        # Initialize the dataloader
        self.init_dataloader(input_layer_key=input_layer_key, preprocess=preprocess, train_frac=self.config.train_frac, use_sampler=True)
        # Initialize the trainer
        self.init_trainer()
        # Train the model
        self.train(save_model=save_model)
        # Reinitialize the dataloader without using the sampler
        self.init_dataloader(input_layer_key=input_layer_key, preprocess=preprocess, train_frac=1.0, use_sampler=False)
        
        # Predict and store the results
        encoded, decoded, class_preds, class_probs, class_true, latent_matrices = self.predict(self.loader, 
                                                                                               return_decoded=return_decoded, decoder_domain=decoder_domain,
                                                                                               return_latent=return_latent,
                                                                                               return_class=return_class, return_class_prob=return_class_prob)
        self.adata.obsm[output_key] = encoded
        if return_decoded:
            if decoder_domain is not None:
                save_key = f"{output_key}_decoded_{decoder_domain}"
            else:
                save_key = f"{output_key}_decoded"
            self.adata.layers[save_key] = decoded
        if return_latent:
            for key, val in latent_matrices.items():
                self.adata.obsm[output_key+'_'+key] = val
        if class_true is not None:
            self.adata.obs[output_key+'_class_true'] = class_true
        if class_preds is not None:
            self.adata.obs[output_key+'_class_pred'] = class_preds
        if class_probs is not None:
            class_probs.index = self.adata.obs.index
            for col in class_probs.columns:
                self.adata.obs[f'class_prob_{col}'] = class_probs[col]

    def get_domain_embeddings(self):
        """
        Retrieves domain embeddings from the trained model.

        Returns:
            pd.DataFrame: A dataframe containing domain embeddings.
        """
        unique_domain_categories = self.adata.obs[self.config.domain_key].cat.categories.values
        domain_labels = torch.tensor(range(len(unique_domain_categories)), dtype=torch.long).to(self.config.device)
        domain_embeddings = self.model.domain_embedding(domain_labels)
        domain_embeddings = domain_embeddings.cpu().detach().numpy()
        domain_df = pd.DataFrame(domain_embeddings, index=unique_domain_categories)
        return domain_df
    
    def get_covariate_embeddings(self):
        """
        Retrieves covariate embeddings from the trained model.

        Returns:
            dict: A dictionary of DataFrames, each containing embeddings for a covariate.
        """
        covariate_dfs = {}
        for covariate_key in self.config.covariate_embedding_dims.keys():
            if covariate_key in self.model.covariate_embeddings:
                unique_covariate_categories = self.adata.obs[covariate_key].cat.categories.values
                covariate_labels = torch.tensor(range(len(unique_covariate_categories)), dtype=torch.long).to(self.config.device)
                covariate_embeddings = self.model.covariate_embeddings[covariate_key](covariate_labels)
                covariate_embeddings = covariate_embeddings.cpu().detach().numpy()
                covariate_df = pd.DataFrame(covariate_embeddings, index=unique_covariate_categories)
                covariate_dfs[covariate_key] = covariate_df
        return covariate_dfs

    def save_model(self, model, save_path):
        """
        Saves the trained model to a file.

        Args:
            model (torch.nn.Module): The trained model.
            save_path (str or Path): Path to save the model file.

        Returns:
            None
        """
        torch.save(model.state_dict(), save_path)
        logger.info(f"Model saved to {save_path}")




    






import os
from ..plotting.pl_enrichment import plot_go_enrichment


def compute_go(feature_list, organism="human", top_n=10, qval_correct=1e-10, color_palette='viridis_r', font_size=12, dpi=300, figsize=(10,3), save_path=None):
    """
    Performs Gene Ontology (GO) enrichment analysis using gseapy.enrichr and visualizes the results.

    Args:
        feature_list (list): List of gene symbols to analyze.
        organism (str, optional): Organism name (e.g., "human", "mouse"). Defaults to "human".
        top_n (int, optional): Number of top enriched GO terms to plot. Defaults to 10.
        qval_correct (float, optional): Maximum adjusted p-value for filtering significant terms. Defaults to 1e-10.
        color_palette (str, optional): Color palette for the enrichment plot. Defaults to 'viridis_r'.
        font_size (int, optional): Font size for plot text. Defaults to 12.
        dpi (int, optional): Resolution of the output plot. Defaults to 300.
        figsize (tuple, optional): Figure size for the plot (width, height). Defaults to (10,3).
        save_path (str, optional): Path to save the plot. If None, the plot is not saved. Defaults to None.

    Returns:
        dict: Enrichment results from `gseapy.enrichr`.
    """

    try:
        import gseapy as gp
    except ImportError:
        raise ImportError("gseapy is required for this method. Please install it using 'pip install gseapy'.")

    gp_results = gp.enrichr(gene_list=feature_list, gene_sets='GO_Biological_Process_2021', organism=organism, outdir=None)
    plot_go_enrichment(gp_results, top_n=top_n, qval_correct=qval_correct, color_palette=color_palette, font_size=font_size, figsize=figsize, dpi=dpi, save_path=save_path)
    return gp_results

def run_gsea_for_all_neurons(ranked_lists, gene_sets='GO_Biological_Process_2021', outdir='GSEA_results',
                             processes = 4, permutation_num=500, seed=0):
    """
    Runs Gene Set Enrichment Analysis (GSEA) for multiple neuron types and saves results.

    Args:
        ranked_lists (dict): Dictionary where keys are neuron names and values are ranked gene lists.
        gene_sets (str, optional): Name of the gene set database to use (e.g., 'GO_Biological_Process_2021'). Defaults to 'GO_Biological_Process_2021'.
        outdir (str, optional): Directory to save GSEA results. Defaults to 'GSEA_results'.
        processes (int, optional): Number of parallel processes for GSEA computation. Defaults to 4.
        permutation_num (int, optional): Number of gene set permutations for statistical significance. Defaults to 500.
        seed (int, optional): Random seed for reproducibility. Defaults to 0.

    Returns:
        dict: A dictionary where keys are neuron names and values are their respective GSEA results.
    """
    try:
        import gseapy as gp
    except ImportError:
        raise ImportError("gseapy is required for this method. Please install it using 'pip install gseapy'.")

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    all_gsea_results = {}

    for neuron_name, ranked_list in ranked_lists.items():
        print(f"Running GSEA for {neuron_name}...")
        neuron_outdir = os.path.join(outdir, neuron_name.replace(" ", "_"))
        if not os.path.exists(neuron_outdir):
            os.makedirs(neuron_outdir)

        # Run GSEA
        gsea_results = gp.prerank(
            rnk=ranked_list,
            gene_sets=gene_sets,
            processes=processes,
            permutation_num=permutation_num,
            outdir=neuron_outdir,
            format='png',
            seed=seed,
            min_size=10,
            max_size=1000,
        )

        all_gsea_results[neuron_name] = gsea_results

    return all_gsea_results


def get_gsea_tables(all_gsea_results):
    """
    Extracts the GSEA summary tables from the enrichment results.

    Args:
        all_gsea_results (dict): Dictionary where keys are neuron names and values are GSEA results.

    Returns:
        dict: A dictionary where keys are neuron names and values are their respective `res2d` DataFrame containing enrichment statistics.
    """
    res_tbls = {}

    for neuron_name, gsea_result in all_gsea_results.items():
        res_tbls[neuron_name] = gsea_result.res2d

    return res_tbls



import torch
from torch.utils.data import Sampler
import logging
import numpy as np
logger = logging.getLogger(__name__)


class ConcordSampler(Sampler):
    """
    A custom PyTorch sampler that performs probabilistic domain-aware and 
    neighborhood-aware batch sampling for contrastive learning.

    This sampler selects samples from both intra-domain and inter-domain 
    distributions based on configurable probabilities.

    Attributes:
        batch_size (int): Number of samples per batch.
        p_intra_knn (float): Probability of selecting samples from k-NN neighborhoods.
        p_intra_domain_dict (dict): Dictionary mapping domain indices to intra-domain probabilities.
        device (torch.device): Device to store tensors (default: GPU if available).
        domain_ids (torch.Tensor): Tensor containing domain labels for each sample.
        neighborhood (Neighborhood): Precomputed k-NN index.
        unique_domains (torch.Tensor): Unique domain categories.
        domain_counts (torch.Tensor): Number of samples per domain.
        valid_batches (list): List of precomputed valid batches.
        min_batch_size (int): Minimum allowed batch size.
    """
    def __init__(self, batch_size, domain_ids, 
                 neighborhood, p_intra_knn=0.3, p_intra_domain_dict=None, min_batch_size=4, device=None):
        """
        Initializes the ConcordSampler.

        Args:
            batch_size (int): Number of samples per batch.
            domain_ids (torch.Tensor): Tensor of domain labels for each sample.
            neighborhood (Neighborhood): Precomputed k-NN index.
            p_intra_knn (float, optional): Probability of selecting samples from k-NN neighborhoods. Default is 0.3.
            p_intra_domain_dict (dict, optional): Dictionary mapping domain indices to intra-domain probabilities. Default is None.
            min_batch_size (int, optional): Minimum allowed batch size. Default is 4.
            device (torch.device, optional): Device to store tensors. Defaults to GPU if available.
        """
        self.batch_size = batch_size
        self.p_intra_knn = p_intra_knn
        print(f"p_intra_knn: {p_intra_knn}")
        self.p_intra_domain_dict = p_intra_domain_dict
        self.device = device or torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.domain_ids = domain_ids
        self.neighborhood = neighborhood
        
        self.unique_domains, self.domain_counts = torch.unique(self.domain_ids, return_counts=True)

        #self.valid_batches,_ = self._generate_batches()
        self.valid_batches = None
        self.min_batch_size = min_batch_size

    # Function to permute non- -1 values and push -1 values to the end
    @staticmethod
    def permute_nonneg_and_fill(x, ncol):
        """
        Permutes non-negative values and fills remaining positions with -1.

        Args:
            x (torch.Tensor): Input tensor containing indices.
            ncol (int): Number of columns to keep.

        Returns:
            torch.Tensor: Permuted tensor with -1s filling unused positions.
        """
        result = torch.full((x.size(0), ncol), -1, dtype=x.dtype, device=x.device)
        for i in range(x.size(0)):
            non_negatives = x[i][x[i] >= 0]
            permuted_non_negatives = non_negatives[torch.randperm(non_negatives.size(0))]
            count = min(ncol, permuted_non_negatives.size(0))
            result[i, :count] = permuted_non_negatives[:count]
        return result


    def _generate_batches(self):
        """
        Generates batches based on intra-domain and intra-neighborhood probabilities.

        Returns:
            list: A list of valid batches.
        """
        all_batches = []

        for domain in self.unique_domains:
            domain_indices = torch.where(self.domain_ids == domain)[0]
            out_domain_indices = torch.where(self.domain_ids != domain)[0]

            # # Test repeating
            # def resample_through_repeat_and_shuffle(indices, nrepeat=10):
            #     len_indices = len(indices)
            #     indices = indices.repeat(nrepeat)
            #     indices = indices[torch.randperm(len(indices))[:len_indices]]
            #     return indices

            # domain_indices = resample_through_repeat_and_shuffle(domain_indices)
            # out_domain_indices = resample_through_repeat_and_shuffle(out_domain_indices)

            # Determine p_intra_domain for the current domain
            p_intra_domain = self.p_intra_domain_dict.get(domain.item())
            num_batches_domain = max(1,(self.domain_counts[domain] // self.batch_size).item()) # number of batches per domain be proportional to domain cell counts
            
            # Sample within knn neighborhood if p_intra_knn > 0
            if self.p_intra_knn == 0:
                batch_global_in_domain_count = int(p_intra_domain * self.batch_size)
                batch_global_out_domain_count = self.batch_size - batch_global_in_domain_count
            else:
                core_samples = domain_indices[torch.randperm(len(domain_indices))[:num_batches_domain]]
                knn_around_core = self.neighborhood.get_knn(core_samples) # (core_samples, k), contains core + knn around the core samples
                knn_around_core = torch.tensor(knn_around_core).to(self.device)
                knn_domain_ids = self.domain_ids[knn_around_core] # (core_samples, k), shows domain of each knn sample
                domain_mask = knn_domain_ids == domain # mask indicate if sample is in current domain
                knn_in_domain = torch.where(domain_mask, knn_around_core, torch.tensor(-1, device=self.device))
                knn_out_domain = torch.where(~domain_mask, knn_around_core, torch.tensor(-1, device=self.device))

                batch_knn_count = int(self.p_intra_knn * self.batch_size)
                batch_knn_in_domain_count = int(p_intra_domain * batch_knn_count)
                batch_knn_out_domain_count = batch_knn_count - batch_knn_in_domain_count
                batch_global_in_domain_count = int(p_intra_domain * (self.batch_size - batch_knn_count))
                batch_global_out_domain_count = self.batch_size - batch_knn_count - batch_global_in_domain_count

                #print(f"batch_knn_count: {batch_knn_count}, batch_knn_in_domain_count: {batch_knn_in_domain_count}, batch_knn_out_domain_count: {batch_knn_out_domain_count}, batch_global_in_domain_count: {batch_global_in_domain_count}, batch_global_out_domain_count: {batch_global_out_domain_count}")

                batch_knn_in_domain = ConcordSampler.permute_nonneg_and_fill(knn_in_domain, batch_knn_in_domain_count)
                batch_knn_out_domain = ConcordSampler.permute_nonneg_and_fill(knn_out_domain, batch_knn_out_domain_count)
                batch_knn = torch.cat((batch_knn_in_domain, batch_knn_out_domain), dim=1) 

            # Sample globally to fill in rest of batch (or all of batch if p_intra_knn == 0)
            if len(domain_indices) < num_batches_domain * batch_global_in_domain_count:
                batch_global_in_domain = domain_indices[
                    torch.randint(len(domain_indices), (num_batches_domain * batch_global_in_domain_count,))].view(num_batches_domain, batch_global_in_domain_count)
            else:
                batch_global_in_domain = domain_indices[
                    torch.randperm(len(domain_indices))[:num_batches_domain * batch_global_in_domain_count]].view(num_batches_domain, batch_global_in_domain_count)

            if len(out_domain_indices) < num_batches_domain * batch_global_out_domain_count:
                batch_global_out_domain = out_domain_indices[
                    torch.randint(len(out_domain_indices), (num_batches_domain * batch_global_out_domain_count,))].view(
                    num_batches_domain, batch_global_out_domain_count)
            else:
                batch_global_out_domain = out_domain_indices[
                    torch.randperm(len(out_domain_indices))[:num_batches_domain * batch_global_out_domain_count]].view(
                    num_batches_domain, batch_global_out_domain_count)

            batch_global = torch.cat((batch_global_in_domain, batch_global_out_domain), dim=1)
            # print shape of batch_knn and batch_global, also print the number of non -1s in each batch/row (sum of non -1s in each row)
            # print(f"batch_knn.shape: {batch_knn.shape}, batch_global.shape: {batch_global.shape}")
            # print(f"batch_knn non -1s: {torch.sum(batch_knn != -1, dim=1)}, batch_global non -1s: {torch.sum(batch_global != -1, dim=1)}")

            sample_mtx = torch.cat((batch_knn, batch_global), dim=1) if self.p_intra_knn > 0 else batch_global

            for _,batch in enumerate(sample_mtx):
                valid_batch = batch[batch >= 0]
                # print length of valid_batch
                # print(f"valid_batch length: {len(valid_batch)}")
                all_batches.append(valid_batch)


        # Shuffle all batches to ensure random order of domains
        indices = torch.randperm(len(all_batches)).tolist()
        all_batches = [all_batches[i] for i in indices]

        return all_batches
    

    def __iter__(self):
        """
        Iterator for sampling batches.

        Yields:
            torch.Tensor: A batch of sample indices.
        """
        self.valid_batches = self._generate_batches()
        for batch in self.valid_batches:
            yield batch

    def __len__(self):
        """
        Returns the number of batches.

        Returns:
            int: Number of valid batches.
        """
        return len(self.valid_batches)



# class ConcordMatchNNSampler(Sampler):
#     def __init__(self, batch_size, indices, domain_ids, 
#                  neighborhood, p_intra_knn=0.3, p_intra_domain_dict=None, 
#                  min_batch_size=4, device=None):
#         # Use half the batch size for the base sampler
#         self.base_sampler = ConcordSampler(
#             batch_size // 2, indices, domain_ids, neighborhood, 
#             p_intra_knn, p_intra_domain_dict, min_batch_size=min_batch_size, device=device)
#         self.neighborhood = neighborhood
#         self.device = device or torch.device('cuda' if torch.cuda.is_available() else 'cpu')

#     def __iter__(self):
#         for batch in self.base_sampler:
#             batch = batch.to(self.device)
#             # Get the first nearest neighbor for each sample, excluding the sample itself
#             nn_indices = self.neighborhood.get_knn(batch, k=1, include_self=False)
#             nn_indices = torch.tensor(nn_indices.squeeze(1), device=self.device)
#             # Concatenate batch and nn_indices to form the full batch
#             full_batch = torch.cat([batch, nn_indices], dim=0)
#             yield full_batch




from .. import logger
from .palettes import get_color_mapping
import pandas as pd
import warnings

# Not used, scanpy has a faster version
def visualize_knn_graph(
    adata, 
    basis='X_pca', 
    dim=2,
    color_by=None, 
    cmap='viridis',
    k=10, 
    layout='spring', 
    node_size=10,
    draw_edges=False, 
    edge_color='gray',
    edge_width=0.1,
    plotly=False,
    ax=None,
    seed=42
):
    """
    Visualizes a k-NN graph from an AnnData object using matplotlib or Plotly.

    Parameters:
    - adata: AnnData object
        The data object containing the data and metadata.
    - basis: str, default='X_pca'
        The basis in adata.obsm, adata.X, or adata.layers for graph computation.
    - color_by: str, default='depth'
        Key in adata.obs to color nodes.
    - k: int, default=5
        Number of neighbors for k-NN graph.
    - layout: str, default='spring'
        Layout algorithm for visualization ('spring', 'spectral', or 'circular').
    - draw_edges: bool, default=True
        Whether to draw edges in the graph.
    - plotly_3d: bool, default=False
        If True, use a 3D Plotly visualization, otherwise use matplotlib.
    - ax: matplotlib.axes._subplots.AxesSubplot, default=None
        Matplotlib axis object to plot on. If None, a new figure is created.
    """
    import networkx as nx
    import numpy as np
    from sklearn.neighbors import NearestNeighbors
    import matplotlib.pyplot as plt
    import plotly.graph_objects as go

    # Get data based on the chosen basis
    if basis in adata.obsm:
        data = adata.obsm[basis]
    elif basis in adata.layers:
        data = adata.layers[basis]
    elif basis == 'X':
        data = adata.X
    else:
        raise ValueError(f"Invalid basis '{basis}', choose a basis from adata.obsm, adata.X, or adata.layers")

    # Build k-NN graph
    logger.info(f"Building {k}-NN graph using {basis} basis")
    nn = NearestNeighbors(n_neighbors=k).fit(data)
    distances, indices = nn.kneighbors(data)
    G = nx.Graph()
    for i, neighbors in enumerate(indices):
        for neighbor in neighbors:
            if i != neighbor:
                G.add_edge(i, neighbor)

    logger.info(f"Graph has {G.number_of_nodes()} nodes and {G.number_of_edges()} edges")

    # Set node colors based on `color_by`
    if color_by is not None:
        if color_by in adata.obs:
            data_col, cmap, palette = get_color_mapping(adata, color_by, {color_by: cmap})
            
            # Handle color mapping for numeric and categorical data
            if pd.api.types.is_numeric_dtype(data_col):
                node_colors = data_col
                is_numeric = True
            else:
                data_col = data_col.astype('category')
                node_colors = data_col.map({cat: palette[i] for i, cat in enumerate(data_col.cat.categories)})
                cmap = None
                is_numeric = False
        else:
            raise ValueError(f"Key '{color_by}' not found in adata.obs")
    else:
        node_colors = np.zeros(adata.shape[0])
        is_numeric = True

    # Assign color as node attribute
    for i, color in enumerate(node_colors):
        G.nodes[i]['color'] = color

    node_color_values = np.array([G.nodes[v]['color'] for v in G.nodes])

    # Choose layout
    logger.info(f"Computing {layout} layout for visualization")
    if layout == 'spring':
        pos = nx.spring_layout(G, dim=dim, seed=seed)
    elif layout == 'spectral':
        pos = nx.spectral_layout(G, dim=dim)
    elif layout == 'circular':
        pos = nx.circular_layout(G, dim=dim)
    elif layout == 'forceatlas2':
        pos = nx.forceatlas2_layout(G, dim=dim, seed=seed)
    elif layout == 'kamada_kawai':
        pos = nx.kamada_kawai_layout(G, dim=dim)
    else:
        raise ValueError(f"Unsupported layout '{layout}'")

    logger.info(f"Plotting {dim}D visualization")
    # 2D Visualization with Matplotlib
    if dim == 2:
        node_positions = np.array([pos[v] for v in G.nodes()])
        
        # Use provided ax if available, otherwise create a new figure
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 8))
        
        # Draw nodes
        sx = ax.scatter(
            node_positions[:, 0], node_positions[:, 1],
            c=node_color_values, cmap=cmap, s=node_size
        )
        
        # Draw edges if enabled
        if draw_edges:
            for i, j in G.edges():
                ax.plot(
                    [pos[i][0], pos[j][0]], [pos[i][1], pos[j][1]],
                    color=edge_color, linewidth=edge_width
                )

        if is_numeric:
            plt.colorbar(sx, ax=ax, label=color_by)
        else:
            ax.legend(handles=[plt.Line2D([0], [0], marker='o', color='w', markerfacecolor=palette[i], markersize=10, label=cat) for i, cat in enumerate(data_col.cat.categories)], title=color_by)

        ax.set_title(f"{k}-Nearest Neighbors Graph ({layout} layout)")
        ax.axis('off')

        # Show only if no ax was supplied
        if ax is None:
            plt.show()
    
    # 3D Visualization with Plotly
    elif dim == 3:
        if not plotly:
            node_positions = np.array([pos[v] for v in G.nodes()])
    
            # Set up the 3D plot
            if ax is None:
                fig = plt.figure(figsize=(10, 8))
                ax = fig.add_subplot(111, projection='3d')
            
            # Draw nodes
            sc = ax.scatter(
                node_positions[:, 0], node_positions[:, 1], node_positions[:, 2],
                c=node_color_values, 
                cmap=cmap, 
                s=node_size
            )
            
            # Draw edges if enabled
            if draw_edges:
                for i, j in G.edges():
                    ax.plot(
                        [pos[i][0], pos[j][0]], 
                        [pos[i][1], pos[j][1]], 
                        [pos[i][2], pos[j][2]], 
                        color=edge_color, linewidth=edge_width
                    )
            
            # Add color bar
            if ax is None:
                cbar = fig.colorbar(sc, ax=ax, shrink=0.5, aspect=5)
                cbar.set_label(color_by)
            
            # Set plot title and turn off the grid for a cleaner look
            ax.set_title(f"{k}-Nearest Neighbors Graph ({layout} layout)")
            ax.grid(False)
            
            if ax is None:
                plt.show()

        else:
            # 3D Plotly visualization remains unchanged since Plotly handles axes differently
            node_xyz = np.array([pos[v] for v in G.nodes()])
            edge_x, edge_y, edge_z = [], [], []
            if draw_edges:
                for i, j in G.edges():
                    edge_x += [pos[i][0], pos[j][0], None]
                    edge_y += [pos[i][1], pos[j][1], None]
                    edge_z += [pos[i][2], pos[j][2], None]
            
            # Create node scatter plot
            node_trace = go.Scatter3d(
                x=node_xyz[:, 0], y=node_xyz[:, 1], z=node_xyz[:, 2],
                mode='markers',
                marker=dict(
                    size=node_size,
                    color=node_color_values,
                    colorscale=cmap,
                    colorbar=dict(title=color_by),
                ),
                text=[f"Node {v}" for v in G.nodes()],
                hoverinfo='text'
            )
            
            # Create edge scatter plot
            if draw_edges:
                edge_trace = go.Scatter3d(
                    x=edge_x, y=edge_y, z=edge_z,
                    mode='lines',
                    line=dict(color=edge_color, width=edge_width),
                    hoverinfo='none'
                )
                fig = go.Figure(data=[edge_trace, node_trace])
            else:
                fig = go.Figure(data=[node_trace])

            fig.update_layout(
                title=f"{k}-Nearest Neighbors Graph ({layout} layout)",
                showlegend=False,
                scene=dict(
                    xaxis=dict(showbackground=False),
                    yaxis=dict(showbackground=False),
                    zaxis=dict(showbackground=False)
                )
            )
            fig.show()
    else:
        raise ValueError("Only 2D and 3D visualizations are supported")






def plot_graph(adata, basis, 
               k=15, layout='kk', seed=42,
               color_by=None, edges=False, edges_width=0.1, edges_color='grey',
                   pal=None, highlight_indices=None,
                   highlight_size=20, draw_path=False, alpha=0.9, text_alpha=0.5,
                   figsize=(9, 3), dpi=300, ncols=1,
                   font_size=8, point_size=10, legend_loc='on data', save_path=None):
    import scanpy as sc
    import matplotlib.pyplot as plt
    import numpy as np

    warnings.filterwarnings('ignore')

    if color_by is None or len(color_by) == 0:
        color_by = [None]  # Use a single plot without coloring

    nrows = int(np.ceil(len(color_by) / ncols))
    fig, axs = plt.subplots(nrows, ncols, figsize=figsize, dpi=dpi, constrained_layout=True)
    axs = np.atleast_2d(axs).flatten()  # Ensure axs is a 1D array for easy iteration

    # Compute knn graph and set graph layout
    sc.pp.neighbors(adata, n_neighbors=k, use_rep=basis, random_state=seed)
    sc.tl.draw_graph(adata, layout=layout, random_state=seed)
    
    if not isinstance(pal, dict):
        pal = {col: pal for col in color_by}

    for col, ax in zip(color_by, axs):
        data_col, cmap, palette = get_color_mapping(adata, col, pal)

        if col is None:
            ax = sc.pl.draw_graph(adata, color=col, ax=ax, show=False,
                                 legend_loc='right margin', legend_fontsize=font_size,
                                 size=point_size, alpha=alpha, edges=edges, edges_width=edges_width, edges_color=edges_color)
        elif pd.api.types.is_numeric_dtype(data_col):
            sc.pl.draw_graph(adata, color=col, ax=ax, show=False,
                            legend_loc='right margin', legend_fontsize=font_size,
                            size=point_size, alpha=alpha, cmap=cmap, edges=edges, edges_width=edges_width, edges_color=edges_color)
        else:
            sc.pl.draw_graph(adata, color=col, ax=ax, show=False,
                            legend_loc=legend_loc, legend_fontsize=font_size,
                            size=point_size, alpha=alpha, palette=palette, edges=edges, edges_width=edges_width, edges_color=edges_color)

        if legend_loc == 'on data':
            for text in ax.texts:
                text.set_alpha(text_alpha)

        # Highlight selected points, check if needs fixing
        if highlight_indices is not None:
            highlight_data = adata[highlight_indices, :]
            if pd.api.types.is_numeric_dtype(data_col):
                sc.pl.embedding(highlight_data, basis=basis, color=col, ax=ax, show=False,
                                legend_loc=None, legend_fontsize=font_size,
                                size=highlight_size, alpha=1.0, cmap=cmap)
            else:
                sc.pl.embedding(highlight_data, basis=basis, color=col, ax=ax, show=False,
                                legend_loc=None, legend_fontsize=font_size,
                                size=highlight_size, alpha=1.0, palette=palette)

            if draw_path:
                embedding = adata.obsm[basis]
                if not isinstance(embedding, np.ndarray):
                    embedding = np.array(embedding)
                path_coords = embedding[highlight_indices, :]
                ax.plot(path_coords[:, 0], path_coords[:, 1], 'r-', linewidth=2)  # Red line for the path

        ax.set_title(ax.get_title(), fontsize=font_size)
        ax.set_xlabel(ax.get_xlabel(), fontsize=font_size)
        ax.set_ylabel(ax.get_ylabel(), fontsize=font_size)

        if hasattr(ax, 'collections') and len(ax.collections) > 0:
            cbar = ax.collections[-1].colorbar
            if cbar is not None:
                cbar.ax.tick_params(labelsize=font_size)

    for ax in axs[len(color_by):]:
        ax.axis('off')

    plt.show()

    if save_path is not None:
        fig.savefig(save_path, dpi=dpi)


# 
def plot_paga(adata, basis, 
            k=15, groups='leiden', resolution=1.0, 
            threshold=0.1,
            layout='kk',
            seed=42,
            color_by=None, 
            node_size_scale=1,
            edge_width_scale=1,
            pal=None, 
            figsize=(9, 3), dpi=300, ncols=1,
            font_size=8, save_path=None):
    import scanpy as sc
    import matplotlib.pyplot as plt
    import numpy as np

    warnings.filterwarnings('ignore')

    if color_by is None or len(color_by) == 0:
        color_by = [None]  # Use a single plot without coloring

    nrows = int(np.ceil(len(color_by) / ncols))
    fig, axs = plt.subplots(nrows, ncols, figsize=figsize, dpi=dpi, constrained_layout=True)
    axs = np.atleast_2d(axs).flatten()  # Ensure axs is a 1D array for easy iteration

    # Compute knn graph and set graph layout
    sc.pp.neighbors(adata, n_neighbors=k, use_rep=basis, random_state=seed)
    
    if groups =='leiden':
        sc.tl.leiden(adata, resolution=resolution)
    elif groups == 'louvain':
        sc.tl.louvain(adata, resolution=resolution)
    else:
        if groups not in adata.obs:
            raise ValueError(f"Group key '{groups}' not found in adata.obs")
    
    sc.tl.paga(adata, groups=groups)
    
    if not isinstance(pal, dict):
        pal = {col: pal for col in color_by}

    for col, ax in zip(color_by, axs):
        data_col, cmap, _ = get_color_mapping(adata, col, pal)

        if col is None:
            sc.pl.paga(adata, threshold=threshold, color=col, ax=ax, show=False,
                            layout=layout, fontsize=font_size,
                            cmap=cmap, node_size_scale=node_size_scale, edge_width_scale=edge_width_scale)
        elif pd.api.types.is_numeric_dtype(data_col):
            sc.pl.paga(adata, threshold=threshold, color=col, ax=ax, show=False,
                        layout=layout, fontsize=font_size,
                        cmap=cmap, node_size_scale=node_size_scale, edge_width_scale=edge_width_scale, colorbar=False)
        else:
            sc.tl.paga(adata, groups=groups) # Strange that this needs to be run right before otherwise it doesn't work
            sc.pl.paga(adata, threshold=threshold, color=col, ax=ax, show=False,
                            layout=layout, fontsize=font_size,
                            cmap=cmap, node_size_scale=node_size_scale, edge_width_scale=edge_width_scale)

        ax.set_title(ax.get_title(), fontsize=font_size)
        ax.set_xlabel(ax.get_xlabel(), fontsize=font_size)
        ax.set_ylabel(ax.get_ylabel(), fontsize=font_size)


    for ax in axs[len(color_by):]:
        ax.axis('off')

    plt.show()

    if save_path is not None:
        fig.savefig(save_path, dpi=dpi)





def compute_paga_layout(adata, groupby_key, weight_threshold=0.3, spring_k=None, seed=42):
    """
    Precompute the PAGA graph layout positions with filtered edges.

    Parameters:
    - adata: AnnData object
    - groupby_key: str, column in adata.obs used to group cells (e.g., 'leiden_Concord')
    - weight_threshold: float, minimum weight for edges to be included in the graph
    - spring_k: float, spring constant for layout positioning
    - seed: int, random seed for reproducibility

    Returns:
    - filtered_graph: NetworkX graph object
    - pos: Dictionary of positions for the nodes
    """
    import networkx as nx

    # Extract the PAGA adjacency matrix
    paga_connectivities = adata.uns['paga']['connectivities']

    # Convert to list of edges with weights
    edges = [
        (i, j, w)
        for i, j, w in zip(*paga_connectivities.nonzero(), paga_connectivities.data)
    ]

    # Filter edges by weight threshold
    filtered_edges = [(i, j, w) for i, j, w in edges if w >= weight_threshold]

    # Build the filtered graph
    filtered_graph = nx.Graph()
    filtered_graph.add_weighted_edges_from(filtered_edges)

    # Relabel nodes to use group labels
    label_mapping = {i: str(label) for i, label in enumerate(adata.obs[groupby_key].cat.categories)}
    filtered_graph = nx.relabel_nodes(filtered_graph, label_mapping)

    # Precompute layout positions
    pos = nx.spring_layout(filtered_graph, weight='weight', k=spring_k, seed=seed)
    return filtered_graph, pos


def plot_paga_custom(adata, filtered_graph, pos, meta_attribute_key, groupby_key, node_size=500, with_labels=True, figsize=(10, 10), pal=None, save_path=None):
    """
    Plot a PAGA graph using a shared layout and colored based on a meta attribute.

    Parameters:
    - adata: AnnData object
    - filtered_graph: Precomputed NetworkX graph
    - pos: Dictionary of shared node positions
    - meta_attribute_key: str, column in adata.obs for node coloring
    - groupby_key: str, column in adata.obs used to group cells
    - node_size: int, size of the nodes in the plot
    - with_labels: bool, whether to display node labels
    - figsize: tuple, size of the figure
    - save_path: str, optional path to save the figure
    """    
    import pandas as pd
    import networkx as nx
    import matplotlib.pyplot as plt
    from matplotlib.colors import Normalize, to_hex
    from . import get_color_mapping
    # Step 1: Assign the meta attribute as a node attribute
    meta_attribute = adata.obs.groupby(groupby_key)[meta_attribute_key].first()
    meta_attribute_dict = meta_attribute.to_dict()
    nx.set_node_attributes(filtered_graph, {node: meta_attribute_dict.get(node, 'Unknown') for node in filtered_graph.nodes}, name=meta_attribute_key)

    # Step 2: Handle color mapping
    if pd.api.types.is_numeric_dtype(adata.obs[meta_attribute_key]):
        # Continuous colormap for numeric values
        values = [meta_attribute_dict[node] for node in filtered_graph.nodes]
        norm = Normalize(vmin=min(values), vmax=max(values))
        data_col, cmap, palette = get_color_mapping(adata, meta_attribute_key, pal=pal)
        node_colors = [to_hex(cmap(norm(meta_attribute_dict[node]))) for node in filtered_graph.nodes]
    else:
        # Discrete colormap for categorical values
        unique_values = meta_attribute.unique()
        data_col, cmap, palette = get_color_mapping(adata, meta_attribute_key, pal=pal)
        node_colors = [palette[filtered_graph.nodes[node][meta_attribute_key]] for node in filtered_graph.nodes]

    # Step 3: Plot the graph
    plt.figure(figsize=figsize)

    # Node labels are the meta attributes
    node_labels = {node: filtered_graph.nodes[node][meta_attribute_key] for node in filtered_graph.nodes} if with_labels else None

    nx.draw(
        filtered_graph,
        pos,
        with_labels=with_labels,
        labels=node_labels,
        node_color=node_colors,
        edge_color='gray',
        node_size=node_size,
        font_size=10
    )

    plt.title(f"PAGA Graph Colored by '{meta_attribute_key}'")
    if save_path:
        plt.savefig(save_path)
    
    plt.show()





# Plot benchmarking results using plottable

def add_metric_row(df):
    """
    Adds a row to a multi-index DataFrame with the first element of each tuple 
    in the multi-index as values in a new 'Metric' row, and keeps only the 
    second element of the multi-index as column headers.
    
    Parameters:
        df (pd.DataFrame): A DataFrame with a multi-index in columns.
        
    Returns:
        pd.DataFrame: Modified DataFrame with a new 'Metric' row.
    """
    import pandas as pd
    df = df.copy()
    # Extract the first and second levels of the multi-index columns
    metric_labels = [col[0] for col in df.columns]
    new_columns = [col[1] for col in df.columns]
    
    # Create a new row with the metric labels
    metric_row = pd.DataFrame([metric_labels], columns=new_columns, index=["Metric"])
    
    # Update the columns of the original DataFrame to only have the second level
    df.columns = new_columns
    
    # Concatenate the metric row at the top of the original DataFrame
    result_df = pd.concat([metric_row, df], axis=0)
    
    return result_df


def plot_benchmark_table(df, pal='PRGn', pal_agg='YlGnBu', cmap_method='norm', cmap_padding=0.05, agg_name='Aggregate score', dpi=300, save_path=None, figsize=None):
    """
    Plots a benchmarking results table using the `plottable` library.

    This function creates a formatted table displaying different benchmarking metrics 
    across various methods. It includes:
    - Circle-marked metric values.
    - Color-encoded values based on a chosen colormap.
    - Aggregate scores visualized as bar charts.

    Args:
        df (pd.DataFrame): 
            The benchmarking results DataFrame. It should have a multi-index in columns 
            where the first level represents metric categories and the second level contains 
            metric names.
        pal (str, optional): 
            Colormap for individual metric values. Defaults to `'PRGn'`.
        pal_agg (str, optional): 
            Colormap for aggregate scores. Defaults to `'YlGnBu'`.
        cmap_method (str, optional): 
            Method for normalizing colormaps. Options:
                - `'norm'`: Normalize based on standard deviation.
                - `'minmax'`: Normalize based on the min-max range.
                - `'minmax_padded'`: Adds padding to the min-max normalization.
                - `'0_to_1'`: Normalize between 0 and 1.
            Defaults to `'norm'`.
        cmap_padding (float, optional): 
            Padding factor for `minmax_padded` colormap normalization. Defaults to `0.05`.
        agg_name (str, optional): 
            The name of the aggregate score column. Defaults to `'Aggregate score'`.
        dpi (int, optional): 
            Resolution of the saved figure. Defaults to `300`.
        save_path (str, optional): 
            If provided, saves the figure to the specified path. Defaults to `None`.
        figsize (tuple, optional): 
            Figure size `(width, height)`. If `None`, it is determined dynamically based on 
            the number of columns and rows. Defaults to `None`.

    Raises:
        ValueError: If `cmap_method` is not one of `'norm'`, `'minmax'`, `'minmax_padded'`, or `'0_to_1'`.

    Returns:
        None
    """
    
    # Plot the geometry results using plotable
    from plottable import ColumnDefinition, Table
    from plottable.plots import bar
    from plottable.cmap import normed_cmap
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    import pandas as pd

    df = df.copy()
    df = add_metric_row(df)
    num_embeds = df.shape[0]
    plot_df = df.drop("Metric", axis=0)
    plot_df['Method'] = plot_df.index

    cmap = mpl.cm.get_cmap(pal)
    cmap_agg = mpl.cm.get_cmap(pal_agg)
    if cmap_method == 'norm':
        cmap_fn = lambda col_data: normed_cmap(col_data, cmap=cmap, num_stds=2.5)
        cmap_agg_fn = lambda col_data: normed_cmap(col_data, cmap=cmap_agg, num_stds=2.5)
    elif cmap_method == 'minmax':
        cmap_fn = lambda col_data: mpl.cm.ScalarMappable(
            norm=mpl.colors.Normalize(vmin=col_data.min(), vmax=col_data.max()),
            cmap=cmap
        ).to_rgba
        cmap_agg_fn = lambda col_data: mpl.cm.ScalarMappable(
            norm=mpl.colors.Normalize(vmin=col_data.min(), vmax=col_data.max()),
            cmap=cmap_agg
        ).to_rgba

    elif cmap_method == 'minmax_padded':
        cmap_fn = lambda col_data: mpl.cm.ScalarMappable(
            norm=mpl.colors.Normalize(vmin=col_data.min()-cmap_padding*(col_data.max()-col_data.min()), vmax=col_data.max()+cmap_padding*(col_data.max()-col_data.min())),
            cmap=cmap
        ).to_rgba
        cmap_agg_fn = lambda col_data: mpl.cm.ScalarMappable(
            norm=mpl.colors.Normalize(vmin=col_data.min()-cmap_padding*(col_data.max()-col_data.min()), vmax=col_data.max()+cmap_padding*(col_data.max()-col_data.min())),
            cmap=cmap_agg
        ).to_rgba
    elif cmap_method == '0_to_1':
        cmap_fn = lambda col_data: mpl.cm.ScalarMappable(
            norm=mpl.colors.Normalize(vmin=0, vmax=1),
            cmap=cmap
        ).to_rgba
        cmap_agg_fn = lambda col_data: mpl.cm.ScalarMappable(
            norm=mpl.colors.Normalize(vmin=0, vmax=1),
            cmap=cmap_agg
        ).to_rgba
    else:
        raise ValueError(f"Invalid cmap_method: {cmap_method}, choose 'norm' or 'minmax'")
    
    column_definitions = [
        ColumnDefinition("Method", width=1.5, textprops={"ha": "left", "weight": "bold"}),
    ]

    aggr_cols = df.columns[df.loc['Metric'] == agg_name]
    stats_cols = df.columns[df.loc['Metric'] != agg_name]

    # Circles for the metric values
    column_definitions += [
        ColumnDefinition(
            col,
            title=col.replace(" ", "\n"),
            width=1,
            textprops={
                "ha": "center",
                "bbox": {"boxstyle": "circle", "pad": 0.1},
            },
            cmap=cmap_fn(plot_df[col]),
            group=df.loc['Metric', col],
            formatter="{:.2f}",
        )
        for i, col in enumerate(stats_cols)
    ]

    column_definitions += [
        ColumnDefinition(
            col,
            width=1,
            title=col.replace(" ", "\n", 1),
            plot_fn=bar,
            plot_kw={
                "cmap": cmap_agg_fn(plot_df[col]),
                "plot_bg_bar": False,
                "annotate": True,
                "height": 0.9,
                "formatter": "{:.2f}",
                # font size
                "textprops": {"fontsize": 12},
            },
            group=df.loc['Metric', col],
            border="left" if i == 0 else None,
        )
        for i, col in enumerate(aggr_cols)
    ]

    # Set figure size dynamically or use provided figsize
    if figsize is None:
        figsize = (len(df.columns) * 1.25, 3 + 0.3 * num_embeds)
    
    with mpl.rc_context({"svg.fonttype": "none"}):
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
        tab = Table(
            plot_df,
            cell_kw={
                "linewidth": 0,
                "edgecolor": "k",
            },
            column_definitions=column_definitions,
            ax=ax,
            row_dividers=True,
            footer_divider=True,
            textprops={"fontsize": 11, "ha": "center"},
            row_divider_kw={"linewidth": 1, "linestyle": (0, (0.5, 2))},
            col_label_divider_kw={"linewidth": 1, "linestyle": "-"},
            column_border_kw={"linewidth": 1, "linestyle": "-"},
            index_col="Method"
        ).autoset_fontcolors(colnames=plot_df.columns)

    if save_path:
        plt.savefig(save_path, bbox_inches="tight")
    plt.show()





import numpy as np
import matplotlib.pyplot as plt

def plot_signal_over_path(sorted_data, sorted_smoothed_data, x, ncols=4, same_yaxis_range=False):
    """
    Plot geodesic distances for each dimension with optional same y-axis range for all plots.

    Parameters:
    sorted_data (ndarray): Original sorted data array.
    sorted_smoothed_data (ndarray): Smoothed data array.
    geodesic_distances (ndarray): Geodesic distances.
    ncols (int): Number of columns in the subplot grid.
    same_yaxis_range (bool): Whether to use the same y-axis range for all plots.

    Returns:
    None
    """
    n_dims = sorted_smoothed_data.shape[1]
    nrows = int(np.ceil(n_dims / ncols))

    # Create the subplots
    fig, axs = plt.subplots(nrows, ncols, figsize=(4 * ncols, 4 * nrows), constrained_layout=True)

    # Ensure axs is a 1D array for easy iteration
    axs = axs.flatten()

    # Determine y-axis range if same_yaxis_range is True
    if same_yaxis_range:
        y_min = np.min([sorted_data.min(), sorted_smoothed_data.min()])
        y_max = np.max([sorted_data.max(), sorted_smoothed_data.max()])

    for i in range(n_dims):
        ax = axs[i]
        ax.plot(x, sorted_data[:, i], color='blue', label='Original Sorted')
        ax.plot(x, sorted_smoothed_data[:, i], color='red', linestyle='--', label='Smoothed')
        ax.set_title(f'Dimension {i}')
        ax.set_xlabel('Geodesic distance to Point A')
        ax.set_ylabel('Value')
        ax.legend()

        # Set the same y-axis range for all plots if required
        if same_yaxis_range:
            ax.set_ylim(y_min, y_max)

    # Turn off any unused axes
    for j in range(n_dims, len(axs)):
        axs[j].axis('off')

    plt.show()





def plot_bar(
    data, key, color='#1f77b4', title=None, 
    order=True, ascending=True, log_scale=False,
    barh=True, legend=False,
    save_path=None, figsize=(4, 3), dpi=300, 
    xlabel_fontsize=8, ylabel_fontsize=8, 
    legend_fontsize=8,
    tick_fontsize=7, title_fontsize=9, bar_width=0.8
):
    import matplotlib.pyplot as plt
    
    fig, ax = plt.subplots(figsize=figsize, dpi=dpi)
    
    # Sort the data based on values in the specified column
    if order:
        data = data.sort_values(by=key, ascending=ascending)
        
    # Plot as horizontal or vertical bar chart
    if barh:
        data.plot(
            kind='barh', y=key, ax=ax, color=color, 
            width=bar_width
        )
        if log_scale:
            ax.set_xscale('log')
    else:
        data.plot(
            kind='bar', y=key, ax=ax, color=color, 
            width=bar_width
        )
        if log_scale:
            ax.set_yscale('log')
    
    # Set title and font sizes
    if title is None:
        title = key
    ax.set_title(title, fontsize=title_fontsize)
    ax.set_xlabel(key, fontsize=xlabel_fontsize)
    ax.set_ylabel(data.index.name if data.index.name else '', fontsize=ylabel_fontsize)
    
    # Set tick font sizes
    ax.tick_params(axis='x', labelsize=tick_fontsize)
    ax.tick_params(axis='y', labelsize=tick_fontsize)
    
    # Legend settings
    if legend:
        ax.legend(
            title=None,
            loc='center left', 
            bbox_to_anchor=(1, 0.5),  # Position to the right, centered vertically
            handletextpad=0.2,
            fontsize=legend_fontsize,
            title_fontsize=legend_fontsize
        )
    else:
        ax.get_legend().remove()
    
    # Save the plot
    if save_path:
        fig.savefig(save_path, dpi=dpi)
        
    plt.show()
    plt.close(fig)





import torch
from torch import nn
from torch.nn import functional as F


# SimCLR contrastive loss function
class nt_xent_loss(nn.Module):
    def __init__(self, temperature=0.5):
        super().__init__()
        self.temperature = temperature

    def forward(self, z_i, z_j):
        batch_size = z_i.shape[0]
        z = torch.cat((z_i, z_j), dim=0)
        z = nn.functional.normalize(z, dim=1)

        similarity_matrix = torch.matmul(z, z.T) / self.temperature
        mask = torch.eye(batch_size * 2, dtype=torch.bool).to(z.device)
        labels = torch.cat([torch.arange(batch_size) for _ in range(2)], dim=0).to(z.device)
        labels = (labels.unsqueeze(0) == labels.unsqueeze(1)).float()
        labels = labels.masked_fill(mask, 0)

        loss = nn.CrossEntropyLoss()(similarity_matrix.masked_fill(mask, -float('inf')), labels.argmax(dim=1))
        return loss


class nt_bxent_loss(nn.Module):
    def __init__(self, temperature=0.5):
        super().__init__()
        self.temperature = temperature

    def forward(self, x, target):
        # Cosine similarity
        xcs = F.cosine_similarity(x[None,:,:], x[:,None,:], dim=-1)
        xcs[torch.eye(x.size(0)).bool()] = float("inf")
        xcs = (xcs / self.temperature).sigmoid()
        loss = F.binary_cross_entropy(xcs, target, reduction="none").to(x.device)
        
        target_pos = target.bool().to(x.device)
        target_neg = ~target_pos
        
        loss_pos = torch.zeros(x.size(0), x.size(0)).to(x.device).masked_scatter(target_pos, loss[target_pos])
        loss_neg = torch.zeros(x.size(0), x.size(0)).to(x.device).masked_scatter(target_neg, loss[target_neg])

        loss_pos = loss_pos.sum(dim=1)
        loss_neg = loss_neg.sum(dim=1)
        num_pos = target.sum(dim=1)
        num_neg = x.size(0) - num_pos

        return ((loss_pos / num_pos) + (loss_neg / num_neg)).mean()



def importance_penalty_loss(importance_weights, penalty_weight=0.1, norm_type='L1'):
    if penalty_weight == 0:
        return torch.tensor(0.0, device=importance_weights.device)

    input_dim = importance_weights.size(0)
    if norm_type == 'L1':
        base_penalty_weight = 1 / input_dim
        penalty_loss = base_penalty_weight * torch.norm(importance_weights, p=1)
    elif norm_type == 'L2':
        base_penalty_weight = 1 / torch.sqrt(torch.tensor(input_dim, dtype=torch.float32))
        penalty_loss = base_penalty_weight * torch.norm(importance_weights, p=2)
    else:
        raise ValueError(f"Unknown norm_type: {norm_type}")

    return penalty_weight * penalty_loss




import pandas as pd
from .. import logger

# Helper function to chunk lists
def chunk_list(lst, chunk_size):
    for i in range(0, len(lst), chunk_size):
        yield lst[i:i + chunk_size]

def get_mouse_genes(human_genes, return_type=None, chunk_size=500):
    try:
        from gseapy import Biomart
    except ImportError:
        raise ImportError("gseapy is required for this method. Please install it using 'pip install gseapy'.")

    bm = Biomart()
    results = []

    total_processed = 0

    for chunk in chunk_list(human_genes, chunk_size):
        h2m_chunk = bm.query(dataset='hsapiens_gene_ensembl',
                             filters={'external_gene_name': chunk},
                             attributes=['external_gene_name', 'mmusculus_homolog_associated_gene_name'])
        results.append(h2m_chunk)
        total_processed += len(chunk)
        logger.info(f"Processed {total_processed} human genes to mouse orthologs.")

    h2m = pd.concat(results, ignore_index=True)

    if return_type == 'dict':
        return dict(zip(h2m['external_gene_name'], h2m['mmusculus_homolog_associated_gene_name']))
    elif return_type == 'pandas':
        return h2m
    else:
        return h2m['mmusculus_homolog_associated_gene_name'].dropna().unique().tolist()

def get_human_genes(mouse_genes, return_type=None, chunk_size=100):
    try:
        from gseapy import Biomart
    except ImportError:
        raise ImportError("gseapy is required for this method. Please install it using 'pip install gseapy'.")
    bm = Biomart()
    results = []
    total_processed = 0

    for chunk in chunk_list(mouse_genes, chunk_size):
        m2h_chunk = bm.query(dataset='mmusculus_gene_ensembl',
                             filters={'external_gene_name': chunk},
                             attributes=['external_gene_name', 'hsapiens_homolog_associated_gene_name'])
        results.append(m2h_chunk)
        total_processed += len(chunk)
        logger.info(f"Processed {total_processed} mouse genes to human orthologs.")

    m2h = pd.concat(results, ignore_index=True)

    if return_type == 'dict':
        return dict(zip(m2h['external_gene_name'], m2h['hsapiens_homolog_associated_gene_name']))
    elif return_type == 'pandas':
        return m2h
    else:
        return m2h['hsapiens_homolog_associated_gene_name'].dropna().unique().tolist()


def get_mouse_genes_offline(human_genes, orthologs, return_type=None, keep_all=False):
    """
    Map human genes to mouse orthologs using the provided ortholog table.

    Parameters:
        human_genes (list): List of human gene symbols to map.
        orthologs (pd.DataFrame): The loaded ortholog table.
        return_type (str): Output format ('dict', 'pandas', or None for a list of unique mouse genes).
        keep_all (bool): If True, return all mapped mouse genes for each human gene.
                                If False, return only the first mapped mouse gene.

    Returns:
        dict, pd.DataFrame, or list: Mouse gene orthologs in the specified format.
    """
    # Filter rows for human and mouse genes
    df_human = orthologs[orthologs["Common Organism Name"] == "human"]
    df_mouse = orthologs[orthologs["Common Organism Name"] == "mouse, laboratory"]

    # Merge human and mouse gene data on 'DB Class Key'
    human_to_mouse = df_human.merge(
        df_mouse,
        on="DB Class Key",
        suffixes=("_human", "_mouse")
    )

    # Filter for human genes of interest
    mapping = human_to_mouse[human_to_mouse["Symbol_human"].isin(human_genes)][["Symbol_human", "Symbol_mouse"]]

    # Group and apply keep_all logic
    if keep_all:
        grouped = mapping.groupby("Symbol_human")["Symbol_mouse"].apply(list)
    else:
        grouped = mapping.groupby("Symbol_human")["Symbol_mouse"].first()

    # Return in the desired format
    if return_type == "dict":
        return grouped.to_dict()
    elif return_type == "pandas":
        return grouped.reset_index()
    else:
        return grouped.explode().unique().tolist()


def get_human_genes_offline(mouse_genes, orthologs, return_type=None, keep_all=False):
    """
    Map mouse genes to human orthologs using the provided ortholog table.

    Parameters:
        mouse_genes (list): List of mouse gene symbols to map.
        orthologs (pd.DataFrame): The loaded ortholog table.
        return_type (str): Output format ('dict', 'pandas', or None for a list of unique human genes).
        keep_all (bool): If True, return all mapped human genes for each mouse gene.
                                If False, return only the first mapped human gene.

    Returns:
        dict, pd.DataFrame, or list: Human gene orthologs in the specified format.
    """
    # Filter rows for human and mouse genes
    df_human = orthologs[orthologs["Common Organism Name"] == "human"]
    df_mouse = orthologs[orthologs["Common Organism Name"] == "mouse, laboratory"]

    # Merge human and mouse gene data on 'DB Class Key'
    mouse_to_human = df_mouse.merge(
        df_human,
        on="DB Class Key",
        suffixes=("_mouse", "_human")
    )

    # Filter for mouse genes of interest
    mapping = mouse_to_human[mouse_to_human["Symbol_mouse"].isin(mouse_genes)][["Symbol_mouse", "Symbol_human"]]

    # Group and apply keep_all logic
    if keep_all:
        grouped = mapping.groupby("Symbol_mouse")["Symbol_human"].apply(list)
    else:
        grouped = mapping.groupby("Symbol_mouse")["Symbol_human"].first()

    # Return in the desired format
    if return_type == "dict":
        return grouped.to_dict()
    elif return_type == "pandas":
        return grouped.reset_index()
    else:
        return grouped.explode().unique().tolist()




import torch
from torch.utils.data import Dataset
from scipy.sparse import issparse
import numpy as np
import logging
logger = logging.getLogger(__name__)


class AnnDataset(Dataset):
    """
    A PyTorch Dataset class for handling annotated datasets (AnnData).

    This dataset is designed to work with single-cell RNA-seq data stored in 
    AnnData objects. It extracts relevant features, domain labels, class labels, 
    and covariate labels while handling sparse and dense matrices.

    Attributes:
        adata (AnnData): The annotated data matrix.
        input_layer_key (str): The key to retrieve input features from `adata`.
        domain_key (str): The key in `adata.obs` specifying domain labels.
        class_key (str, optional): The key in `adata.obs` specifying class labels.
        covariate_keys (list, optional): A list of keys for covariate labels in `adata.obs`.
        device (torch.device): The device to store tensors (GPU or CPU).
        data (torch.Tensor): Tensor containing input data.
        domain_labels (torch.Tensor): Tensor containing domain labels.
        class_labels (torch.Tensor, optional): Tensor containing class labels if provided.
        covariate_tensors (dict): A dictionary containing tensors for covariate labels.
        indices (np.ndarray): Array of dataset indices.
        data_structure (list): A list describing the dataset structure.
    """
    def __init__(self, adata, input_layer_key='X', domain_key='domain', class_key=None, covariate_keys=None, device=None):
        """
        Initializes the AnnDataset.

        Args:
            adata (AnnData): The annotated dataset.
            input_layer_key (str, optional): Key to extract input data. Defaults to 'X'.
            domain_key (str, optional): Key for domain labels in `adata.obs`. Defaults to 'domain'.
            class_key (str, optional): Key for class labels in `adata.obs`. Defaults to None.
            covariate_keys (list, optional): List of keys for covariate labels in `adata.obs`. Defaults to None.
            device (torch.device, optional): Device to store tensors (GPU or CPU). Defaults to GPU if available.

        Raises:
            ValueError: If domain or class key is not found in `adata.obs`.
        """
        self.adata = adata
        self.input_layer_key = input_layer_key
        self.domain_key = domain_key
        self.class_key = class_key
        self.covariate_keys = covariate_keys if covariate_keys is not None else []
        self.device = device or torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        self.data = torch.tensor(self._get_data_matrix(), dtype=torch.float32)
        self.domain_labels = torch.tensor(self.adata.obs[self.domain_key].cat.codes.values, dtype=torch.long).to(
            self.device)
        self.indices = np.arange(len(self.adata))

        if self.class_key:
            self.class_labels = torch.tensor(self.adata.obs[self.class_key].cat.codes.values, dtype=torch.long).to(
                self.device)
        else:
            self.class_labels = None

        self.covariate_tensors = {
            key: torch.tensor(self.adata.obs[key].cat.codes.values, dtype=torch.long).to(self.device)
            for key in self.covariate_keys
        }

        self.data_structure = self._init_data_structure()

        logger.info(f"Initialized dataset with {len(self.indices)} samples. Data structure: {self.data_structure}")

    def _get_data_matrix(self):
        """
        Retrieves the feature matrix from `adata`.

        Returns:
            np.ndarray: The feature matrix as a NumPy array.

        Raises:
            KeyError: If the specified input layer is not found.
        """
        if self.input_layer_key == 'X':
            return self.adata.X.toarray() if issparse(self.adata.X) else self.adata.X
        else:
            return self.adata.layers[self.input_layer_key].toarray() if issparse(self.adata.layers[self.input_layer_key]) else \
            self.adata.layers[self.input_layer_key]

    def get_embedding(self, embedding_key, idx):
        """
        Retrieves embeddings for a given key and index.

        Args:
            embedding_key (str): The embedding key in `adata.obsm`.
            idx (int or list): Index or indices to retrieve.

        Returns:
            np.ndarray: The embedding matrix.

        Raises:
            ValueError: If the embedding key is not found.
        """
        if embedding_key == 'X':
            return self.adata.X.toarray()[idx]
        elif embedding_key in self.adata.obsm.key():
            return self.adata.obsm[embedding_key][idx]
        else:
            raise ValueError(f"Embedding key '{embedding_key}' not found in adata")

    def get_domain_labels(self, idx):
        """
        Retrieves the domain labels for a given index.

        Args:
            idx (int or list): Index or indices to retrieve.

        Returns:
            torch.Tensor: The domain labels.
        """
        if self.domain_labels is not None:
            return self.domain_labels[idx]
        return None

    def get_class_labels(self, idx):
        """
        Retrieves the class labels for a given index.

        Args:
            idx (int or list): Index or indices to retrieve.

        Returns:
            torch.Tensor: The class labels.
        """
        if self.class_labels is not None:
            return self.class_labels[idx]
        return None

    def _init_data_structure(self):
        """
        Initializes the structure of the dataset.

        Returns:
            list: A list defining the dataset structure.
        """
        structure = ['input']
        if self.domain_key is not None:
            structure.append('domain')
        if self.class_key is not None:
            structure.append('class')
        structure.extend(self.covariate_keys)
        structure.append('idx')
        return structure

    def get_data_structure(self):
        """
        Returns the data structure of the dataset.

        Returns:
            list: A list defining the dataset structure.
        """
        return self.data_structure

    def __len__(self):
        """
        Returns the number of samples in the dataset.

        Returns:
            int: The dataset size.
        """
        return len(self.indices)

    def __getitem__(self, idx):
        """
        Retrieves the dataset items for the given index.

        Args:
            idx (int or list): The sample index.

        Returns:
            tuple: A tuple containing input data, domain labels, class labels, and covariate tensors.
        """
        if isinstance(idx, torch.Tensor):
            idx = idx.cpu().numpy()
        actual_idx = self.indices[idx]
        data_tensor = self.data[actual_idx]
        items = []
        for key in self.data_structure:
            if key == 'input':
                items.append(data_tensor)
            elif key == 'domain' and self.domain_labels is not None:
                items.append(self.domain_labels[actual_idx])
            elif key == 'class':
                items.append(self.class_labels[actual_idx])
            elif key in self.covariate_keys:
                items.append(self.covariate_tensors[key][actual_idx])
            elif key == 'idx':
                items.append(torch.tensor(actual_idx))

        return tuple(items)

    def shuffle_indices(self):
        """
        Shuffles dataset indices.
        """
        np.random.shuffle(self.indices)

    def subset(self, idx):
        """
        Creates a subset of the dataset with the given indices.

        Args:
            idx (list): Indices of the subset.

        Returns:
            AnnDataset: A new AnnDataset instance containing only the selected indices.
        """
        # Create a new AnnDataset with only the selected idx
        subset_adata = self.adata[idx].copy()
        return AnnDataset(subset_adata, self.input_layer_key, self.domain_key, self.class_key, self.covariate_keys, self.device)




import logging
import random
from pathlib import Path
import numpy as np
import torch
import pandas as pd
from . import io


def set_seed(seed):
    """set random seed."""
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    # if n_gpu > 0:
    #     torch.cuda.manual_seed_all(seed)




def add_file_handler(logger: logging.Logger, log_file_path: Path):
    """
    Add a file handler to the logger.
    """
    h = logging.FileHandler(log_file_path)

    # format showing time, name, function, and message
    formatter = logging.Formatter(
        "%(asctime)s-%(name)s-%(levelname)s-%(funcName)s: %(message)s",
        datefmt="%H:%M:%S",
    )
    h.setFormatter(formatter)
    h.setLevel(logger.level)
    logger.addHandler(h)


def natural_key(string_):
    import re
    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_)]


def sort_string_list(string_list):
    return sorted(string_list, key=natural_key)

